/* $Header:   P:/source/cwright/cwstart/c.c_v   1.75   13 Dec 1994 10:00:28   ericj  $ */
/*
 *
 *                Copyright 1991,1992,1993 Starbase Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Starbase Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Starbase Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Starbase Corp.
 *
 *    Owner:
 *        Starbase Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */

//#include "win_filt.h"
#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

/*** Definitions for 'sas.c' ***/
void DLL  _sas_init( void );
void DLL  _sas_routines( void );
void DLL  _sas_indent_nl( void );
BOOL DLL  _sas_insert_brace( void );


LIBMAIN    /* defines the LibMain() */


#define LIGHT

#define MAX_BUF_LEN    100

static LPVOID _sas_hash_names        = NIL;  /* pointer to the hash table containing */
                                           /* all 'SAS' keyword names. */
static LPVOID _sas_hash_names_1      = NIL;  /* pointer to the hash table containing */
                                           /* user defined keyword names. */
static int _sas_hash_size_1          = 47;
static LPVOID _sas_hash_names_2      = NIL;  /* pointer to the hash table containing */
                                           /* user defined keyword names. */
static int _sas_hash_size_2          = 47;

static LPVOID _sas_hash_templates    = NIL;  /* pointer to the hash table containing */
                                           /* an abbreviated name and its template */

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _sas_hash_templates.
 */
typedef struct _LanguageTemplate {         
	LPSTR abbrev;
	LPSTR tpl;
} LanguageTemplate;


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR C_INDENT_STR = "\377if\377while\377else\377do\377for\377switch\377";

/*
 * _sas_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _sas_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _sas_templates[] = {
	{ "if",                         "if (&)" },
	{ "ifb",                        "if (&)%0{\n}" },
	{ "wh|whi|whil|while",          "while (&)%0{\n}" },
	{ "do",                         "do%0{\n%1&\n%2}\nwhile( );" },
	{ "fo|for" ,                    "for (&; ; )%0{\n}" },
	{ "{",                          "%0{\n%1&\n%2}" },
	{ "sw|swi|swit|switc|switch",   "switch (&)%0{\n}" },
	{ "br|bre|brea|break",          "break;\n" },
	{ "co|con|cont|conti|contin|continu|continue",      
                                    "continue;\n" },
	{ "ca|cas|case",                "case &:\n\tbreak;" },
	{ "de|def|defa|defau|defaul|default",      
                                    "default:\n\t&\nbreak;" },
	{ "el|els|else",                "else%0{\n%1&\n%2}" },
	{ "ei",                         "else if (&)%0{\n}" },
	{ "re|ret|retu|retur|return",   "return(&);\n" },
	{ NIL,       NIL },
};

/* Define the list of 'SAS' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
LPSTR _sas_names[] = {
	// SAS statements used in the DATA step pg.iii
	"ABORT",
	"ARRAY",
	"ATTRIB",
	"BY",
	"CALL",
	"CARDS",
	"CARDS4",
	"DATA",
	"DELETE",
	"DO",
	"DROP",
	"END",
	"ERROR",
	"FILE",
	"FORMAT",
	"GOTO",
	"IF", "THEN", "ELSE"
	"INFILE",
	"INFORMAT",
	"INPUT",
	"KEEP",
	"LABEL",
	"LENGTH",
	"LINK",
	"LIST",
	"LOSTCARD",
	"MERGE",
	"MISSING",
	"OUTPUT",
	"PUT",
	"RENAME",
	"RETAIN",
	"RETURN",
	"SELECT",
	"SET",
	"STOP",
	"UPDATE",
	// SAS expressions pg. 223
	"EQ",
	"NE",
	"GT",
	"LT",
	"GE",
	"LE",
	"AND",
	"OR",
	"NOT",
	// SAS functions pg. 232 (selected)
	"ABS",
	"DIM",
	"MAX",
	"MIN",
	"MOD",
	"SIGN",
	"SQRT",
	"CEIL",
	"FLOOR",
	"FUZZ",
	"INT",
	"ROUND",
	"EXP",
	"LOG",
	"SUM",
	"NORMAL",
	"RANUNI",
	"LAG",
	// SAS statements used in DATA steps and PROC's pg. iv (selected)
	"ENDSAS",
	"RUN",
	"TITLE",
	// SAS statements used in PROC's pg. 335
	"ATTRIB",
	"BY",
	"CLASS",
	"FORMAT",
	"FREQ",
	"ID",
	"LABEL",
	"MODEL",
	"OUTPUT",
	"PARMCARDS",
	"PROC",
	"WEIGHT",
	"VAR",
	// Others
	"INFILE",
	"OPTIONS",
	"PAGE",
	"LIBNAME",
	"FILENAME",
	"OUT",
	NIL,
};


static long _sas_in_comment( long offset );
static long _sas_find_comment_start( void );
static long _sas_find_comment_end( UINT style );
static BOOL _sas_in_string( long offset );

void DLL _sas_init( void );
BOOL DLL _sas_expand( void );
long DLL _sas_language_colors_update( long firstline, long lastline );
void DLL _sas_language_colors( long numlines );
void DLL _sas_assign_template( LPSTR abbrev, LPSTR tpl );
void DLL _sas_add_keyword( LPSTR keyword, int table );
BOOL DLL _sas_indent( void );

static int lastChar(LPSTR str,long pos);
static int firstChar(LPSTR str,long pos);
static LPSTR ReadLine(void);
static void TabMove(int ctab);


void DLL
_init( void )
{
	_sas_init();
}

/*
 ** _sas_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expandion
 *  and keyword coloring.
 */
void DLL
_sas_init( void )
{
	static BOOL firstTime = TRUE;
	int    i;

	if (firstTime)
	{
		LibExport( "BOOL _sas_expand");
		LibExport( "_sas_assign_template LPSTR LPSTR");
		LibExport( "_sas_add_keyword LPSTR int");
		LibExport( "BOOL _sas_indent( void ) ");
		LibExport( "BOOL _sas_insert_brace");
		LibExport( "long _sas_language_colors_update long long");
		LibExport( "_sas_language_colors long");
		LibExport( "_sas_routines" );
		LibExport( "_sas_hash_size int int" );

		firstTime = FALSE;

		/* Create a hash table to hold all of the 'SAS' keyword names.
		 */
		_sas_hash_names = HashCreateTable( 101 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;

		while ( _sas_names[ i ] != NIL )
		{
			LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _sas_hash_names, 
								_sas_names[i], strlen( _sas_names[i] ) );
			if (lpf)
			{
				*lpf = StrNew( _sas_names[i] );
				i++;
			}
		}

		/* Create a hash table to hold all of the 'SAS' template names.
		 */
		_sas_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), HASH_IGN_CASE );
		i = 0;

		while ( _sas_templates[ i ].abbrev != NIL )
		{
			_sas_assign_template( _sas_templates[i].abbrev, _sas_templates[i].tpl );
			i++;
		}
	}
}

void DLL
_sas_hash_size( int ht, int size )
{
	if (size > 0)
	{
		if (ht == 1)
			_sas_hash_size_1 = size;
		else if (ht == 2)
			_sas_hash_size_2 = size;
	}
}


/*
 ** _sas_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'SAS' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_sas_expand( void )
{
	LPVOID htEntry;
	LPSTR  _sas_string = NULL;
	LPSTR  str;
	int    i = 0;

	MarkSavePos();

	/* Check to see if we are currently positioned inside a 'SAS' comment
	 */
	if (_sas_in_comment( -1L ) != -1)
		goto _sas_expand_incomplete;

	str = BufReadStr( 65535 );
	while (str[i] && !isspace(str[i]))
		i++;
	str[i] = '\0';
	_sas_string = StrTrim( str, " \t" );
	StrFree( str );

	if (!_sas_string || !*_sas_string)
		goto _sas_expand_incomplete;

	/* Search for a keyword match in the hash table.
	 */
	htEntry = HashFindEntry( _sas_hash_templates, _sas_string, strlen( _sas_string ) );

	if (htEntry)
	{
		/* A template exists for the current keyword, delete the keyword
		 * and insert the template.
		 */
		MarkDropPos( 1 );
		str = *(LPSTR*)htEntry;
		BufDelChar( lstrlen( _sas_string ) );

		_ext_expand_template( str );
	} 
	else
		goto _sas_expand_incomplete;

	StrFree( _sas_string );
	return TRUE;

_sas_expand_incomplete:
	StrFree( _sas_string );
	MarkRestorePos();

	return FALSE;
}


/*
 ** _sas_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL
_sas_language_colors_update( long firstline, long lastline )
{
	long startoff;
	long ec_line  = 0; /* end comment line */
	UINT style    = 0;
	int  color;

	if (firstline && lastline)
	{
		/* Begin coloring at the previous comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		if (firstline == 1)
			_PosPrevLine( 0L );
		else
			_PosNextLine( firstline - 1);

		startoff = _PosQOffset();
		if ( _sas_find_comment_start() == 1)
		{
			startoff  = _PosQOffset();

			_PosNextChar();   /* move off of the comment */
			_PosNextChar();

			style = 1;

			if (_sas_find_comment_end( style ))
				ec_line = _PosQLine();
		}

		_PosInit( startoff );
		_PosCurrentCharNL();
		while (_PosPrevCharNL() && (_PosPrevCharNL() == '\\'))
		{
			_PosPrevLine( 0L );
			_PosCurrentCharNL();
			startoff  = _PosQOffset();
		}

		_PosInit( startoff );
		firstline = _PosQLine();

		/* End coloring at the next comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		color = _PosQColor();
		_PosNextLine( lastline - 1 );
		_PosNextLine( 0 );

		if (color || _PosQColor())
		{
			_sas_find_comment_end( style );
			lastline = _PosQLine();
		}
		else if (_sas_find_comment_end( style ) == 2)
			lastline = _PosQLine();


		SysSetSystemFlags( SYSTEM_DELAY_ATTR_CHANGES );

		if (!ec_line)
			AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
		else if (ec_line < lastline)
		{
			AttrSetColor( ec_line, lastline, 1L, 0x7FFFFFFFL, 0 );
			if (ec_line < firstline)
				firstline = ec_line;
		}

		/* Reset the attributes for the line range within _sas_language_colors.
		 */
		_PosInit( startoff );
		_sas_language_colors( lastline - firstline + 1 );

		/* Allow changes to be redraw 
		 */
		SysClrSystemFlags( SYSTEM_DELAY_ATTR_CHANGES );
		WinRedrawLines( WinQCurrentWindow(), firstline, lastline );
	}

	return(lastline - firstline);
}



/*
 ** _sas_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_sas_language_colors( long numlines )
{
	long  startoff;
	long  startline;
	long  endoff;
	UINT  ch;
	UINT  lastch = ';';
	LPSTR buf;
	int   i = 0;
	int   comment_color    = ColorComments( -1 );
	int   keyword_color    = ColorKeywords( -1 );
	long  maxlines         = ExtCommentSearchLimit( -1 );
	int   quote_color      = ColorAlternate1( -1 );
	int   number_color     = ColorAlternate2( -1 );
	int   brace_color      = ColorAlternate3( -1 );
	int   preproc_color    = ColorAlternate4( -1 );


	buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

	/* Start at the top of the buffer or current position 
	 */
	if (numlines == -1)
	{
		_PosInit( 0 );  
		numlines = 0x7FFFFFFF;
	}
	else
	{
		startoff  = _PosQOffset();
		startline = _PosQLine();
		endoff    = _sas_in_comment( startoff );
		if (endoff == -1)
			_PosInit( startoff );
		else
		{
			_PosInit( endoff );
			numlines += (startline - _PosQLine());
		}
	}
	ch = _PosCurrentChar();
	startoff = _PosQOffset();

	maxlines = max( numlines, maxlines );

	while ((ch != EOF_CHAR) && (numlines > 0))
	{
		if (ch == '\n')
			numlines--;

		if (ch == '"' || ch == '\'')
		{
			UINT quote = ch;

            startoff = _PosQOffset();

			do 
			{
				ch = _PosNextChar();
				if (ch == '\\')
				{
					ch = _PosNextChar();
					if ((ch == quote) || (ch == '\\') || (ch == '\n') || (ch == '\r'))
					{
						if (ch == '\n')
							numlines++;
						else if (ch == '\r')
							ch = _PosNextChar();
						ch = _PosNextChar();
					}
				}

				if ((ch == EOF_CHAR) || (ch == '\n'))
					break;
			} while (ch != quote);

			if (ch == quote)
				ch = _PosNextChar();

			/* color the string */
			if (quote_color)
			{
				endoff = _PosQOffset();
	    		_PosInit( startoff );
				_PosSetColor( quote_color, endoff - startoff );
				_PosInit( endoff );
			}
			continue;
		}
		else if (number_color && isdigit( ch ))
		{
            int j = 0;
			UINT lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			startoff = _PosQOffset();
			do
			{
                j++;
				ch = _PosNextChar();
			} while (isxdigit( ch ) 
					|| (ch == '.') 
					|| (((ch == 'x') || (ch == 'X')) && j == 1));

			if ((  lastch != ' ' 
				&& lastch != '\t' 
				&& lastch != '\r' 
				&& lastch != '\n' 
				&& lastch != EOF_CHAR
				&& !ispunct( lastch ))
				|| lastch == '_'
				|| (ch == '_'))
				continue;

            if (ch == 'l' || ch == 'L')
                ch = _PosNextChar();

			endoff = _PosQOffset();
    		_PosInit( startoff );
			_PosSetColor( number_color, endoff - startoff );
			_PosInit( endoff );
			continue;
		}
		else if ( (ch == '%' || ch == '&') && preproc_color)
		{
			startoff = _PosQOffset();
			ch = _PosNextChar();
			if ( ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r')
				continue;

			do
			{
				ch = _PosNextChar();
			} while ( (isalnum(ch) || ch == '_') && ch != EOF_CHAR);
			endoff = _PosQOffset();
    		_PosInit( startoff );
			_PosSetColor( preproc_color, endoff - startoff );
			_PosInit( endoff );
			continue;
		}
		else if (isalnum( (BYTE)ch ) || (ch == '_'))
		{
			UINT lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			i = 0;
			startoff = _PosQOffset();

			do
			{
				if (i < MAX_BUF_LEN)
					buf[i++] = (char)ch;
				ch = _PosNextChar();
			} while (isalnum( (BYTE)ch ) || (ch == '_'));

			if ((lastch != ' ' 
						&& lastch != '\t' 
						&& lastch != '\r' 
						&& lastch != '\n' 
						&& lastch != EOF_CHAR
						&& !ispunct( lastch ))
					|| (ch == '_'))
				i = 0;
			continue;
		}
		else if (i)
		{
			int color = 0;
			if (_sas_hash_names_1 && HashFindEntry( _sas_hash_names_1, buf, i ))
				color = ColorAlternate5( -1 );
			else if (_sas_hash_names_2 && HashFindEntry( _sas_hash_names_2, buf, i ))
				color = ColorAlternate6( -1 );
			else if (HashFindEntry( _sas_hash_names, buf, i ))
				color = keyword_color;

			if (color)
			{
				endoff = _PosQOffset();
				_PosInit( startoff );
				_PosSetColor( color, endoff - startoff );
				_PosInit( endoff );
			}
			i = 0;
		} 
		// look for funny SAS-style comments
		else if (ch == '*' && lastch == ';')
		{
            startoff = _PosQOffset();

			do 
			{
				ch = _PosNextChar();
				if ((ch == EOF_CHAR) || (ch == '\n'))
					break;
			} while (ch != ';');

			if (ch == ';')
				ch = _PosNextChar();
			/* color the comment */
			endoff = _PosQOffset();
    		_PosInit( startoff );
			_PosSetColor( comment_color, endoff - startoff );
			_PosInit( endoff );
			continue;
		}
		else if (ch == '/')
		{
			startoff = _PosQOffset();
			ch = _PosNextChar();

			if (ch == '*')
			{
				long lines = 0;

				ch = _PosNextChar();
				while ((ch != EOF_CHAR) && (lines < maxlines))
				{
					if (ch == '\n')
					{
						lines++;
						numlines--;
					}
					else if (ch == '*')
					{
						ch = _PosNextChar();
						if (ch == '/')
						{
							endoff = _PosQOffset();
							_PosInit( startoff );
							_PosSetColor( comment_color, endoff - startoff + 1 );
							_PosInit( endoff );
							break;
						}
						else
							continue;
					}
					ch = _PosNextChar();
				}
				ch = _PosNextChar();
			}
			continue;
		}
		else if (brace_color && _fstrchr("{}", ch) != 0)    
		{                                        
			long coff;                            

			coff = _PosQOffset();            
			_PosInit(coff);                  
			_PosSetColor(brace_color, 1);    
			_PosInit(coff);                  
		}                                   

		// record last non-whitespace character (for SAS-comment detection)
		if (!isspace(ch))
			lastch = ch;

		ch = _PosNextChar();
	}
	MemFree( buf );
}

void DLL
_sas_add_keyword( LPSTR keyword, int table )
{
	if (keyword && *keyword && table >= 0)
	{
		LPSTR XFAR *lpf;
		LPVOID  ht;

		switch( table )
		{
		case 0:
			ht = _sas_hash_names;
			break;
		case 1:
			if (_sas_hash_names_1 == NULL)
				_sas_hash_names_1 = HashCreateTable( _sas_hash_size_1, sizeof( LPSTR ), HASH_IGN_CASE );
			ht = _sas_hash_names_1;
			break;
		case 2:
			if (_sas_hash_names_2 == NULL)
				_sas_hash_names_2 = HashCreateTable( _sas_hash_size_2, sizeof( LPSTR ), HASH_IGN_CASE );
			ht = _sas_hash_names_2;
			break;
		}

		lpf = (LPSTR XFAR *)HashGetEntry( ht, 
								keyword, strlen( keyword ) );
		if (lpf)
			*lpf = StrNew( keyword );
	}
}


/*
 ** _sas_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    tpl      - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'SAS' template strings
 *    without changing the code.
 */
void DLL
_sas_assign_template( LPSTR abbrev, LPSTR tpl )
{
	_ext_assign_template( _sas_hash_templates, abbrev, tpl );
}


/*
 ** _sas_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'SAS' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in C_INDENT_STR - perform smart
 *       indentation.  This may include skipping the next line if
 *       it consists only of an open brace.
 */
BOOL DLL
_sas_indent( void )
{
	long   curline;
	UINT   endch;

	MarkSavePos();

	curline = BufQCurrentLine();

	/* Check to see if we are currently positioned inside a 'SAS' comment
	 */
	if (_sas_in_comment( -1L ) != -1)
		goto _sas_indent_incomplete;

	/* Check to see if there is a trailing '{' on the current line,
	 * and if so, indent on the next line.
	 */
	do {
		if (MovPrevChar( 1L ) == FALSE)
			break;
		endch = BufReadChar();
	} while (endch == ' ' || endch == '\t');
	MarkRestorePos();
	if (endch == '{')
		goto _sas_indent_next_line;
	MarkSavePos();

	/* Not in a comment so indent according to the current 'SAS' construct.
	 */

	/* Check to see if a '{' is by itself on the next line. If so, skip
	 * it and indent on the line following it.
	 */
	if (SrchFind( "^[ \t]*{\\c[ \t]*$", SEARCH_FORWARD | SEARCH_REGEX, NIL )
			&& (BufQCurrentLine() == curline + 1) )
		MarkDropPos( 1 );
	else
		MarkRestorePos();

_sas_indent_next_line:
	_sas_indent_nl();
	return TRUE;

	_sas_indent_incomplete:
		MarkRestorePos();

	return FALSE;
}


void DLL
_sas_indent_nl( void )
{
	/*
	**	Desc: a newline function that maintains the current indentation level
	**			if cstyle switch is nonzero C style indenting is enabled. The rules
	**			for indenting and outdenting are as follows:
	**				1. If the line ends in ',' and the previous line does
	**					not, then indent one tab ( corresponds to a function call
	**					that is not completed on a line and is broken between two
	**					arguments)
	**				2. If the line ends in ';' and the previous line ends in ','
	**					then outdent one tab. ( correponds to the end of a
	**					function call broken over multiple lines.)
	**
	**				3. If the line ends in '{' then indent one tab
	**
	**				4. If the line ends in '}' then outdent one tab
	**
	**				5. If the line ends in ')' or ':' then indent one tab
	**					(corresponds to a c construct ie. if (), case :, etc.)
	**
	**				6. If the line ends in anything else and starts with '{'
	**					then indent one tab (open block with statement on line)
	**
	**				7. Otherwise align with previous non-blank line.
	**
	**			The above suits my style of formatting for c programs. Feel free
	**			to experiment and modify the above to suit your own style.
	**
	**History:
	** 15-Feb-1990 JVV: Created
	** 01-Oct-1990 JVV: Added modifications for c programming
	** 10-Dec-1990 EOX: Various mods. Improved tab-movement.
	**    Oct-1992 EOX: CodeWright implementation.
	**
	**Author:
	** Erik Oxaal      Company:    Ergosoft Innovasjon AS, OSLO NORWAY Oct 1992
	**
	*/
	int firstch, lastch1, lastch2;
	LPSTR buf = NULL, vbuf = NULL, p;
	LONG xStart, xTarg,
	yStart, yTarg;
	int  brstyle = ExtSetStyle( NULL, -1 );

	/*	 Save current position */

	xStart = BufQCurrentCol();
	yStart = BufQCurrentLine();

	lastch1 = lastch2 = 0;
	/* find start of previous non blank line and collect information on it */

	MovEOL();
	if (SrchFind( "^[ \t]*\\c[^ \t\r\n]+", ~SEARCH_FORWARD & SEARCH_REGEX, NIL ))
	{
		xTarg = BufQCurrentCol();
		yTarg = BufQCurrentLine();

		// Read current line, but make sure leading whitespace consists 
		// only of the correct number of spaces, not tabs.
		buf = ReadLine();
		if (yTarg == yStart)
		{
			lastch1 = lastChar(buf,xStart - 1);
			firstch = firstChar(buf,xStart -1);
		}
		else
		{
			int cch = _fstrlen(buf);
			lastch1 = lastChar(buf,cch);
			firstch = firstChar(buf,cch);
		}

		/* now see if there is a non blank line before this one
		** and get its final non blank character
		*/

		MovAbs(0, 1);
		if (SrchFind( "^[ \t]*\\c[^ \t\r\n]+", ~SEARCH_FORWARD & SEARCH_REGEX, NIL ))
		{
			vbuf = ReadLine();

			if (yTarg != BufQCurrentLine())
				lastch2 = lastChar(vbuf,32767);
		}

		MovAbs(yTarg,xTarg);
		switch(lastch1)
		{
		/* function continuation */

		case ',':
			// Idea for future enhancement: indent all the way to
			// the opening parantheses. If implemented, this should
			// also work correctly in the case of nested function calls
			// and properly disregard parantheses contained in comments,
			// strings, etc. For now, just indent extra tab.
			
			if (lastch2 != ',')
				TabMove(2);
			break;

		/* c construct */

		case ')':
		case ':':
			TabMove(1);
			break;

		/* end of function call or statement */

		case ';':
			if (lastch2 == ',')
				TabMove(-2);
			else if (lastch2 == ')')
				TabMove(-1);
			else if (lastch2 == 'e' && strstr(vbuf, "else"))
				TabMove(-1);
			else if (firstch == '{' && strchr(buf,'}') == NULL)
				TabMove(1);
			break;

		/* block open */

		case '{':
			if ((brstyle == 1) || (brstyle == 2))
				TabMove(1);
			break;

		/* block close */

		case '}':
			if (brstyle == 3)
				TabMove(-1);
			break;

		default:
			/* see if block is opened and not closed on line */

			if ( lastch1 == 'e' && strstr(buf, " else")
					|| lastch1 == 'o' && strstr(buf, " do"))
				TabMove(1);
			else if (firstch == '{' && strchr(buf,'}') == NULL)
				TabMove(1);
		}

		/* determine what column new line should start at */
		xTarg = BufQCurrentCol();
	}
	else
		/* no previous non blank lines, start from first column */
		xTarg = 1;

	if (buf)
		StrFree(buf);
	if (vbuf)
		StrFree(vbuf);
	/* point at new line */
	yTarg = yStart + 1;

	/* Move to original starting position */
	MovAbs(yStart, xStart);

	/* If in insert mode, insert newline and force the start of the next
	** line to the previous starting column. If not in insert mode and the
	** next line is empty, start the line at the previous starting column,
	** otherwise, move to the start of the line.
	*/

	if (BufQInsertMode())
	{
		buf = BufReadStr(32767);
		BufDelToEOL();
		BufInsertEOL();
		MovAbs(yTarg,xTarg);
		BufInsertStr(buf);
		StrFree(buf);
		MovAbs(yTarg,xTarg);
	}
	else
	{
		MovAbs(yTarg,1);
		buf = ReadLine();
		p = buf;

		while (isspace(*p))
		p++;

		if (p == buf + _fstrlen(buf))	// Line empty or just whitespace
			MovAbs(yTarg,xTarg);
		else if (isspace(*buf))
			SrchFind( "[ \t]*\\c[^ \t]+", SEARCH_FORWARD|SEARCH_REGEX, NIL );
		StrFree(buf);
	}
}

BOOL DLL
_sas_insert_brace( void )
{
	LPVOID htEntry;
	int fEmptyLine;
	LPSTR buf;
	long xTarg, yTarg;

	xTarg = BufQCurrentCol();
	yTarg = BufQCurrentLine();

	if ((_sas_in_comment(-1) == -1) &&
		(!_sas_in_string(-1)))
	{

		if (xTarg == 1)
		{ /* function delimiter */
			_ext_expand_template( "{\n%1&\n%home}" );
		}
		else
		{
			MovAbs(0, 1);
			buf = ReadLine();
			fEmptyLine = (firstChar(buf, _fstrlen(buf)) == 0);
			StrFree(buf);
			MovAbs(yTarg, xTarg);
		
			if (fEmptyLine)
				if (MovUp( 1 ))
					MovEOL();
		
			_sas_init();
			htEntry = HashFindEntry( _sas_hash_templates, "{", 1 );

			if (htEntry)
				_ext_expand_template( *(LPSTR*)htEntry );
			else
				BufInsertChar( 0x7b );
		}
	}
	else
		BufInsertChar( 0x7b );

	return TRUE;
}


/*--  Private functions --*/

/**************************************************************************/


static int 
firstChar(LPSTR str,long pos)
{
	/*
	**	Desc: Returns the first non-white character of the string
	**			before position pos or '\0'if none
	**	 Pre: Input string
	**	Post: first char
	** History:
	**	01-Oct-1990 JVV: Created
	*/

	LPSTR pTmp;
	LPSTR endpoint;


	pTmp = str;
	endpoint = str + pos;
	while ( *pTmp && pTmp < endpoint && (*pTmp == ' ' || *pTmp == '\t'))
		pTmp++;

	if (!(*pTmp) || pTmp == endpoint)
		return 0;

	return(*pTmp);
}

/**************************************************************************/

static int 
lastChar(LPSTR str,long pos)
{
	/*
	**	Desc: Finds the last non whitespace character before the position pos
	**	 Pre: Input string and position
	**	Post: last nonblank char
	** History:
	**	01-Oct-1990 JVV: Created
	*/

	LPSTR pTmp;

	pTmp = str + min(((int)pos),(int)(strlen(str) - 1));

	while (pTmp >= str && (*pTmp == ' ' || *pTmp == '\t'))
		pTmp--;

	if (pTmp < str)
		return 0;

	return *pTmp;
}

/**************************************************************************/

// Return contents of current line, starting at current column. Substitute
// spaces for all positions preceding current column.
// History:
//      1992.10.10      EOX     Created

static LPSTR 
ReadLine(void)
{
	LPSTR pLine, p;
	long  colFirst = BufQCurrentCol();

	pLine = StrNNew("", (int)colFirst);
	_fmemset(pLine, ' ', (int)colFirst);
	pLine[colFirst] = 0;
	pLine = StrAppend(pLine, p = BufReadStr(32767));
	StrFree(p);
	return(pLine);
}

/**************************************************************************/

// Move cursor forward or backward 'ctab' tabstops on current line.
// History:
//      1990.12         EOX     Created
//      1992.10.10      EOX     CodeWright version (re-write)

static void 
TabMove(int ctab)
{
	if (ctab > 0)
		while (ctab--)
			MovAbs(0, BufQNextTab(0));
	else
		while (ctab++ < 0)
			MovAbs(0, BufQPrevTab(0));
}

/*
 ** _sas_in_string()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'SAS' string by search backward for a ". 
 *
 *  Returns BOOL
 */
static BOOL
_sas_in_string( long offset )
{
	UINT  ch;
	long  quotes     = 0;

	/* Check to see if we are currently positioned inside a 'SAS' string
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	while (TRUE)
	{
		do
		{
			if (ch == '"' || ch == '\'')
			{
				//ch = _PosPrevChar();
				quotes++;
			}
			ch = _PosPrevChar();
		} while ((ch != EOF_CHAR) && (ch != '\n'));

		if (ch == '\n')
			ch = _PosPrevChar();
	 	if (ch == '\r')
			ch = _PosPrevChar();
		if (ch != '\\')
			break;
		else
			ch = _PosPrevChar();
	}




	return ((int)quotes & 1);
}


/*
 ** _sas_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'SAS' comment by search backward for a // or /* 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_sas_in_comment( long offset )
{
	UINT  ch;
	long  lines     = 0;
	long  maxlines = ExtCommentSearchLimit( -1 );

	/* Check to see if we are currently positioned inside a 'SAS' comment
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	do
	{
		switch( ch )
		{
		case '*':
			ch = _PosPrevChar();
			if (ch == '/')    /* Found the beginning of a comment */
			{
				offset = _PosQOffset();  /* currrent line is in a comment */
				ch = _PosPrevChar();
				if ((ch != '/') || (ch == '/' && (lines==0)))
				{
					_PosInit( offset );
					return offset;
				}
			}
			break;
		case '/':
			ch = _PosPrevChar();
			if (ch == '/')
			{
				if (lines==0)
					return _PosQOffset();  /* currrent line is in a comment */
				else
					return -1;    /* Not in a comment */
			}
			else if (ch == '*')
				return -1L;    /* Not in a comment */
			break;

		case '"':
			ch = _PosPrevChar();
			while (TRUE)
			{
				while (ch != EOF_CHAR && ch != '"' && ch != '\n')
					ch = _PosPrevChar();

				if (ch == '\n')
				{
					lines++;
					ch = _PosPrevChar();
				}
			 	if (ch == '\r')
					ch = _PosPrevChar();
				if (ch != '\\')
					break;
				else
					ch = _PosPrevChar();
			}
			ch = _PosPrevChar();
			break;

		case '\n':
			lines++;
		default:
			ch = _PosPrevChar();
			break;
		}
	} while ((ch != EOF_CHAR) && (lines < maxlines));

	return -1L;
}

/*
 ** _sas_find_comment_start()
 *
 *  Search backwards for a 'SAS' comment sequence. Return 1 if /<*> found
 *  and 2 if <*>/ found, 0 otherwise. Leave the current position
 *  at the beginning of the found comment.
 */
static long 
_sas_find_comment_start( void )
{
	UINT ch    = _PosCurrentChar();
	long  lines = 0;
	long  maxlines = ExtCommentSearchLimit( -1 );

	do
	{
		if (ch == '*')
		{
			ch = _PosPrevChar();
			if (ch == '/')
			{
				long offset = _PosQOffset();  /* currrent line is in a comment */
				ch = _PosPrevChar();
				_PosInit( offset );

				if (ch == '/')
				{
					if (lines == 0)
						break;
				}
				else
					return 1;
			}
		}
		else if (ch == '/')
		{
			ch = _PosPrevChar();
			if ((ch == '/') && (lines == 0))
				break;
			if (ch == '*')
			{
				long offset = _PosQOffset();  /* currrent line is in a comment */
				ch = _PosPrevChar();
				if (ch == '/')
					return 1;
				_PosInit( offset );
				return 2;
			}
		}
		else
		{
			if (ch == '\n')
				lines++;
			else if (ch == '"')
			{
				ch = _PosPrevChar();
				while (TRUE)
				{
					while (ch != EOF_CHAR && ch != '"' && (ch == '\n'))
					{
						if (ch == '\n')
							lines++;
						ch = _PosPrevChar();
					}
					if (ch == '\n')
						ch = _PosPrevChar();
				 	if (ch == '\r')
						ch = _PosPrevChar();
					if (ch != '\\')
						break;
					else
						ch = _PosPrevChar();
				}
			}
			ch = _PosPrevChar();
		}
	} while ((ch != EOF_CHAR) && (lines < maxlines));

	return 0;
}

/*
 ** _sas_find_comment_end()
 *
 *  Search forwards for a 'SAS' comment sequence. Return 1 if /<*> found
 *  and 2 if <*>/ found, 0 otherwise.  Leave the current position
 *  at the end of the found comment.
 */
static long
_sas_find_comment_end( UINT style )
{
	long lines = 0;
	UINT ch    = _PosCurrentChar();
	long  maxlines = ExtCommentSearchLimit( -1 );

	do
	{
		if (ch == '*')
		{
			ch = _PosNextChar();
			if (ch == '/')
				return 2;
		}
		else if (ch == '/')
		{
			ch = _PosNextChar();
//			if ((ch == '/') && style)
//				break;
			if (ch == '*')
				return 1;
		}
		else
		{
			if (ch == '\n')
				lines++;
			else if (ch == '"')
			{
				int eolchars;

				ch = _PosNextChar();
				while (TRUE)
				{
					while (ch != EOF_CHAR && ch != '"' && ch != '\n')
					{
						if (ch == '\n')
							lines++;
						ch = _PosNextChar();
					}
					eolchars = 1;
					if (ch == '\n')
						ch = _PosPrevChar();
				 	if (ch == '\r')
						ch = _PosPrevChar(), eolchars = 2;
					if (ch != '\\')
						break;
					else
					{
						_PosNextChar();
						if (eolchars == 2)
							_PosNextChar();
						ch = _PosNextChar();
					}
				}
			}
			ch = _PosNextChar();
		}
	} while ((ch != EOF_CHAR) && (lines < maxlines));

	return 0;
}





void DLL
_sas_routines( void )
{
	int   matches = 0;
	LPSTR str     = "^[a-zA-Z0-9_].*\\((.|\n)*\\)";
	LPSTR stre    = "^[a-zA-Z0-9_].*\\((.|\n)*\\)[ \t\n]*\\c[^ \t\n]";
#ifndef LIGHT
	long  curline;
	UINT  ch;
	long  lines;
#endif

#ifdef LIGHT
	if (Grep( "^DATA|^data|^PROC|^proc" ))
		MsgMessage( "" );
	else
		MsgMessage( "No routines found." );
#else
	SysWaitStart();

	MarkSavePos();

	BufSetAscii();
	MovTopBuf();

	while (SrchFind( str, SEARCH_REGEX | SEARCH_FORWARD, 0 ))
	{
		curline = BufQCurrentLine();

		SrchFind( stre, SEARCH_REGEX | SEARCH_FORWARD, 0 );

		ch = BufReadChar();
		if (ch != ';')
		{
			if (!matches)
				AttrSetVisible( 0, 0, 0 );
			matches++;
			lines = BufQCurrentLine() - curline
						- !!(ch == '{');
			AttrSetVisible( curline, curline + (lines < 0L ? 0L : lines), 1 );
			MovHome();
		}

		if (!MovDown( 1 ))
			break; 
	}

	if (matches)
	{
		BufSetCompact();
		MarkDropPos( 1 );
		MovTopBuf();
	}
	else
		MarkRestorePos();

	SysWaitEnd();

	MsgMessage( matches ? "" : "No routines found." );
#endif
}
